#pragma once
#include "../FlipullerGB/solver.h"

namespace FlipullistGB{
	public ref struct ConstValue{
		array<int>^ BlockCodes;
		int BlockCodeInvalid;
		array<System::String^>^ BlockStrings;
		System::String^ BlockStringInvalid;
		array<System::String^>^ PosStrings;
		array<System::String^>^ FilenamesBlockTemplate;
		array<System::String^>^ FilenamesNumTemplate;

		ConstValue(){
			// ubN̐̕ݒ
			this->BlockCodes = gcnew array<int>{CELL_EMPTY,CELL_EMPTY,CELL_WALL,CELL_PIPE_V,0,1,2,3,CELL_EXTEND,BLOCK_WILD};
			this->BlockStrings = gcnew array<System::String^>{"@","@","","V","","","","~","r","r"};
			this->BlockCodeInvalid = CELL_INVALID;
			this->BlockStringInvalid = "H";
			// ʒũeLXg
			this->PosStrings = gcnew array<System::String^>{
				"ŏi",
				"\"CLEAR\"̉",
				"\"TIME\"̏",
				"\"TIME\"",
				"\"TIME\"̉",
				"\"BLOCK\"̏",
				"\"BLOCK\"",
				"\"BLOCK\"̉",
				"\"STAGE\"̏",
				"\"STAGE\"",
				"\"STAGE\"̉",
				"ŉi"
			};
			// t@C
			this->FilenamesBlockTemplate = gcnew array<System::String^>{
				"\\empty.bmp",
				"\\arrow.bmp",
				"\\wall.bmp",
				"\\pipe.bmp",
				"\\block_o.bmp",
				"\\block_t.bmp",
				"\\block_r.bmp",
				"\\block_x.bmp",
				"\\block_s.bmp"
			};
			this->FilenamesNumTemplate = gcnew array<System::String^>{
				"\\num_0.bmp",
				"\\num_1.bmp",
				"\\num_2.bmp",
				"\\num_3.bmp",
				"\\num_4.bmp",
				"\\num_5.bmp",
				"\\num_6.bmp",
				"\\num_7.bmp",
				"\\num_8.bmp",
				"\\num_9.bmp"
			};
		}

	};

	public ref class Front{
	internal:
		literal int NumBlocks = 9;
		literal int NumNums = 10;
		literal int FieldWidth = 10;
		literal int FieldHeight = 12;
		literal int FieldPosX = 1;
		literal int FieldPosY = 3;
		literal int QuotaPosX = 16;
		literal int QuotaPosY = 4;
		literal int HoldPosX = 11;
		literal int HoldPosY = 3;
		literal int HoldHeight = 12;
	internal:
		ConstValue Consts;

		GBScreenCapture::GBScrennCaptureRegionForm^ CaptureForm;
		System::Drawing::Bitmap^ CapturedImage;
		array<System::Drawing::Bitmap^>^ BlockTemplates;
		array<System::Drawing::Bitmap^>^ NumTemplates;

		int CurrentHeight;
		int Quota;
		System::String^ BlockText;
		System::String^ HoldText;
		Flipuller::Problem* _Problem;
		Flipuller::Solution* _Solution;
		System::String^ SolutionText;

		Front(){
			this->_Problem = NULL;
			this->_Solution = NULL;
		}
		~Front(){
			if(this->_Problem){delete this->_Problem; this->_Problem = NULL;}
			if(this->_Solution){delete this->_Solution; this->_Solution = NULL;}
		}

		void LoadTemplates(System::String^ dirname){
			this->BlockTemplates = gcnew array<System::Drawing::Bitmap^>(this->NumBlocks);
			this->NumTemplates = gcnew array<System::Drawing::Bitmap^>(this->NumNums);
			// ubN̓ǂݍ
			for(int i = 0; i < this->NumBlocks; i++){
				this->BlockTemplates[i] = gcnew System::Drawing::Bitmap(dirname + this->Consts.FilenamesBlockTemplate[i]);
			}
			// ̓ǂݍ
			for(int i = 0; i < this->NumNums; i++){
				this->NumTemplates[i] = gcnew System::Drawing::Bitmap(dirname + this->Consts.FilenamesNumTemplate[i]);
			}
		}

		void AnalyzeImage(){
			if(this->CapturedImage == nullptr) return;
			if(this->_Problem){delete this->_Problem; this->_Problem = NULL;}
			this->_Problem = new Flipuller::Problem();
			// ڕW
			int quota1 = GBScreenCapture::ImageProcessing::TemplateSelection8x8(
				this->CapturedImage,
				this->QuotaPosX,
				this->QuotaPosY,
				this->NumTemplates);
			int quota0 = GBScreenCapture::ImageProcessing::TemplateSelection8x8(
				this->CapturedImage,
				this->QuotaPosX + 1,
				this->QuotaPosY,
				this->NumTemplates);
			this->Quota = quota1 * 10 + quota0;
			this->_Problem->quota = this->Quota;

			// ubN̔zu
			System::Text::StringBuilder^ sbText = gcnew System::Text::StringBuilder();
			for(int h = 0; h < Front::FieldHeight; h++){
				sbText->Append(h.ToString("D2"));
				sbText->Append(": ");
				for(int w = 0; w < Front::FieldWidth; w++){
					int idx = GBScreenCapture::ImageProcessing::TemplateSelection8x8(
						this->CapturedImage,
						this->FieldPosX + w,
						this->FieldPosY + h,
						this->BlockTemplates);
					if(idx >= 0){
						sbText->Append(this->Consts.BlockStrings[idx]);
						this->_Problem->cell[h][w] = this->Consts.BlockCodes[idx];
					}else{
						sbText->Append(this->Consts.BlockStringInvalid);
						this->_Problem->cell[h][w] = this->Consts.BlockCodeInvalid;
					}
				}
				sbText->AppendLine();
			}
			this->BlockText = sbText->ToString();

			// 茳̃ubN
			this->CurrentHeight = -1;
			this->HoldText = this->Consts.BlockStringInvalid;
			this->_Problem->block = this->Consts.BlockCodeInvalid;
			for(int h = Front::HoldPosY + Front::HoldHeight - 1; h >= Front::HoldPosY; h--){
					int idx = GBScreenCapture::ImageProcessing::TemplateSelection8x8(
						this->CapturedImage,
						this->HoldPosX,
						h,
						this->BlockTemplates);
					// ǂ牽炾疳
					if(idx < 4 ) continue;
					if(idx == 8) idx = 9;
					this->HoldText = this->Consts.BlockStrings[idx];
					this->_Problem->block = this->Consts.BlockCodes[idx];
					this->CurrentHeight = idx;
					break;
			}
		}

		int Solve(){
			if(!this->_Problem) return 0;
			if(this->_Solution){delete this->_Solution; this->_Solution = NULL;}
			this->_Solution = new Flipuller::Solution();

			if(solve(this->_Problem,this->_Solution)){
				System::Text::StringBuilder^ sb = gcnew System::Text::StringBuilder();
				for(int i = 0; i < this->_Solution->num_move; i++){
					// eLXg
					sb->Append((i+1).ToString("D3"));
					sb->Append(": ");
					sb->Append(this->_Solution->move[i].ToString("D2"));
					sb->Append(" (");
					sb->Append(this->Consts.PosStrings[this->_Solution->move[i]]);
					sb->Append(")");
					sb->AppendLine();
				}
				this->SolutionText = sb->ToString();
				return 1;
			}else{
				return 0;
			}
		}
	};
};